"""
For managing the game map & camera
"""

import math

import trg.terminal
from trg.game import Game
from trg.player import Player
from trg.type import Map


def read_map() -> str:
    """
    Reads the map text file (`data/map.txt`)

    MASSIVE WARNING: USE A RECTANGULAR MAP. NON RECTANGULAR MAPS BREAK EVERYHING
    """

    with open("data/map.txt", "r", encoding="UTF-8") as file:
        content = file.read()
        return content.strip("\n")


def parse_map(raw_map: str) -> Map:
    """
    Parses the map
    """

    parsed_map: Map = []
    lines = raw_map.split("\n")

    for line in lines:
        chars = list(line)

        parsed_map.append(chars)

    return parsed_map


def calculate_camera_columns(line: list[str], player: Player, game: Game):
    """
    Calculate what part of the row can fit on the screen

    For example:

    ```
    ----------------------i------------------------
        ^ CHOPS OFF HERE       AND HERE ^
        < -------- TERM SIZE ---------- >
    ```
    Note: This example will probably look very strange if you don't use a monospace font
    """

    term_size = game.TERM_SIZE.columns

    area_from = math.floor(player.coordinates[0] - (term_size / 2))
    area_to = math.floor(player.coordinates[0] + (term_size / 2))

    # This code essentially makes it so that if the player gets close enough to the edge,
    # The camera will not move past the edge. This is pretty common for camera systems.

    # Close to left edge
    if area_from < 0:
        area_from = 0
        area_to = term_size

    # Close to right edge
    if area_to > len(line):
        area_to = len(line)
        area_from = term_size * -1

    area_to_print = line[area_from:area_to]

    return area_to_print


def calculate_camera_rows(game_map: Map, player: Player, game: Game):
    """
    Calculate which rows can fit on the screen

    For example:
    ```

                                      ---
              ^     CHOPS OFF HERE -> ---
              |                       -i-
    TERM SIZE |                       ---
              |                       ---
              .     CHOPS OFF HERE -> ---
                                      ---
    ```
    Note: This example will probably look very strange if you don't use a monospace font
    """

    # -1 To account for trailing newline at the end of the map
    term_size = game.TERM_SIZE.lines - 1

    area_from = math.floor(player.coordinates[1] - (term_size / 2))
    area_to = math.floor(player.coordinates[1] + (term_size / 2))

    # Close to top
    if area_from < 0:
        area_from = 0
        area_to = term_size

    # Close to bottom
    if area_to > len(game_map):
        area_to = len(game_map)
        area_from = term_size * -1

    area_to_print = game_map[area_from:area_to]

    return area_to_print


def render_map(game: Game, player: Player):
    """
    Renders the map and prints it to the screen
    """

    # Make copy of game map
    # Note: any modifications we make to 'map' will also apply to 'game.map' because of refrences
    game_map = game.game_map

    # Save previous tile
    previous_tile = game_map[player.coordinates[1]][player.coordinates[0]]

    # Plot the player
    game_map[player.coordinates[1]][player.coordinates[0]] = "i"

    # Calculate how to print the map
    map_to_print = ""
    area_of_map_to_print = calculate_camera_rows(game_map, player, game)

    for line in area_of_map_to_print:
        area_of_line_to_print = calculate_camera_columns(line, player, game)
        map_to_print += "".join(area_of_line_to_print) + "\n"

    # After it's calculated, we put back the tile where the player use to be,
    # so we don't have duplicates of the player
    # This is actually much faster than making a deep copy of 'game.map',
    # and we need this to be as fast as possible so the terminal doesn't flash
    game_map[player.coordinates[1]][player.coordinates[0]] = previous_tile

    # Clearing after we have already calculated the map to reduce the flashing,
    # so we can just display the old map while the new one is rendering
    trg.terminal.clear()

    # NOTE: This will produce a trailing newline (\n at the end of the output)
    # This is to prevent cursor flicker on some terminals
    print(map_to_print.strip("\n"))



"""
Main file for TRG
"""

from trg.player import Player
from trg.game import Game, Mode

from trg import game_map, terminal, inventory, game_input

def main():
    """
    Main function, only executes if the script is run as a program
    """

    game = Game()
    player = Player()

    print("======================")
    print(" TRG, by Tal & Henri")
    print("======================")
    input("Press enter to begin")

    print("Initializing map...")
    game.game_map = game_map.parse_map(game_map.read_map())
    terminal.clear()

    # Game loop
    while True:
        if game.mode == Mode.MAP:
            game_map.render_map(game, player)
        elif game.mode == Mode.INVENTORY:
            inventory.render_inventory(game, player)
        game_input.read_input(game, player)


if __name__ == "__main__":
    main()



"""
For displaying and managing the inventory menu
"""

from trg.game import Game
from trg.player import Player


def render_inventory(game: Game, player: Player):
    """
    For rendering the inventory menu

    TODO: Not done yet
    """
    print("This is not done yet...")



"""
For managing all items
"""


class Item:
    """
    It's an item, pretty self explanitory
    """

    def __init__(self, name: str):
        self.name = name


class Weapon(Item):
    """
    A weapon, inherits from Item
    """

    def __init__(self, name: str, damage: int):
        super().__init__(name)
        self.damage = damage


Items = {
    "ROCKET_LAUNCHER": Weapon("Rocket Launcher", 4),
}



"""
For capturing user input and then executing code based on that input
"""

import readchar

from trg.game import Mode, Game
from trg.player import Player


def read_input(game: Game, player: Player):
    """
    Reads a singular character and then executes something based on that character

    Keybinds:

    E:          Open inventory
    M:          Open map
    S:          Open shop (TODO)
    ARROW KEYS:
                MAP:       Moves the character
                INVENTORY: Navigates inventory
                BATTLE:    Navigates battle menu (TODO)
                SHOP:      Navigates shop menu   (TODO)
    """

    key = readchar.readkey()

    # Switching Modes
    if key == "e":
        game.mode = Mode.INVENTORY
    if key == "m":
        game.mode = Mode.MAP

    # MAP mode keybinds
    elif game.mode == Mode.MAP:
        if key == readchar.key.UP:
            player.move_up(game)
        elif key == readchar.key.DOWN:
            player.move_down(game)
        elif key == readchar.key.LEFT:
            player.move_left(game)
        elif key == readchar.key.RIGHT:
            player.move_right(game)



"""
Contains the player class & any other player logic
"""

from trg.game import Game

class Player:
    """
    Player class, should only be one at a time!
    """

    coordinates = [0, 0]
    hp = 20
    money = 100
    level = 1
    inventory = []

    def move_up(self, game: Game):
        """
        Moves the player up by reducing `coordinates[1]` by 1
        """

        if self.coordinates[1] <= 0:
            return

        self.coordinates[1] -= 1

    def move_down(self, game: Game):
        """
        Moves the player down by increasing `coordinates[1]` by 1
        """

        if self.coordinates[1] >= len(game.game_map) - 1:
            return

        self.coordinates[1] += 1

    def move_left(self, game: Game):
        """
        Moves the player left by decreasing `coordinates[0]` by 1
        """

        if self.coordinates[0] <= 0:
            return

        self.coordinates[0] -= 1

    def move_right(self, game: Game):
        """
        Moves the player left by increasing `coordinates[0]` by 1
        """

        if self.coordinates[0] >= len(game.game_map[0]) - 1:
            return

        self.coordinates[0] += 1



"""
For managing the game state & information
"""

from enum import Enum
from shutil import get_terminal_size

from trg.type import Map


class Mode(Enum):
    """
    Describes the mode the game is in
    """

    MAP = 1
    INVENTORY = 2
    SHOP = 3
    BATTLE = 4


class Game:
    """
    Game information

    This is a class so that we can actually edit information like the mode
    """

    game_map: Map = []
    TERM_SIZE = get_terminal_size()
    NAME = "TRG"
    mode = Mode.MAP



"""
For managing the terminal
"""

from shutil import get_terminal_size


def clear():
    """
    Clears the screen by printing a lot of new lines
    """

    # Prints out a ton of new lines to clear the terminal
    # Somehow it is faster than os.system("clear") and minimizes flashing
    print("\n" * get_terminal_size().lines, end="\x1b[H")
